/*  RAxML-VI-HPC (version 2.0) a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright June 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  stamatak@ics.forth.gr
 *
 *  When publishing work that is based on the results from RAxML-HPC please cite:
 *
 *  Alexandros Stamatakis, Thomas Ludwig, and Harald Meier: "RAxML-III: A Fast Program for Maximum Likelihood-based Inference of Large Phylogenetic Trees". 
 *  In Bioinformatics 21(4):456-463.
 */


#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h> 
#include <math.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifdef PARALLEL
#include <mpi.h>
#endif

#include "axml.h"
#include "globalVariables.h"

/***************** UTILITY FUNCTIONS **************************/


double gettime()
{
  struct timeval ttime;
  gettimeofday(&ttime , NULL);
  return ttime.tv_sec + ttime.tv_usec * 0.000001;
}

int gettimeSrand()
{
  struct timeval ttime;
  gettimeofday(&ttime , NULL);
  return ttime.tv_sec + ttime.tv_usec;
}

double randum (long  *seed)
    /* random number generator, modified to use 12 bit chunks */
  { /* randum */
    long  sum, mult0, mult1, seed0, seed1, seed2, newseed0, newseed1, newseed2;
    double res;

    mult0 = 1549;
    seed0 = *seed & 4095;
    sum  = mult0 * seed0;
    newseed0 = sum & 4095;
    sum >>= 12;
    seed1 = (*seed >> 12) & 4095;
    mult1 =  406;
    sum += mult0 * seed1 + mult1 * seed0;
    newseed1 = sum & 4095;
    sum >>= 12;
    seed2 = (*seed >> 24) & 255;
    sum += mult0 * seed2 + mult1 * seed1;
    newseed2 = sum & 255;

    *seed = newseed2 << 24 | newseed1 << 12 | newseed0;
    res = 0.00390625 * (newseed2 + 0.000244140625 * (newseed1 + 0.000244140625 * newseed0));  

    return res;   
  } /* randum */

int filexists(char *filename)
{
  FILE *fp;
  int res;
  fp = fopen(filename,"r");
  
  if(fp) 
    {
      res = 1;
      fclose(fp);
    }
  else 
    res = 0;
       
  return res;
} 

/********************* END UTILITY FUNCTIONS ********************/


/******************************some functions for the likelihood computation ****************************/

void *getxnode (nodeptr p)  
{ 
  nodeptr  s;

  if (! (p->x)) 
    { 
      if ((s = p->next)->x || (s = s->next)->x) 
	{
	  p->x = s->x;
	  s->x = NULL;
        }     
    }
  return  p->x;
}




void hookup (nodeptr p, nodeptr q, double z)
  { /* hookup */
    p->back = q;
    q->back = p;
    p->z = q->z = z;
  } /* hookup */

/******************************some functions for the likelihood computation ****************************/




/***********************reading and initializing input ******************/

void getnums (rawdata *rdta)
    /* input number of species, number of sites */
  { /* getnums */
    

    if (fscanf(INFILE, "%d %d", & rdta->numsp, & rdta->sites) != 2) 
      {
	printf("ERROR: Problem reading number of species and sites\n");
	exit(-1);
      }
   

    if (rdta->numsp < 4) 
      {
	printf("TOO FEW SPECIES\n");
	exit(-1);
      }

    if (rdta->sites < 1) 
      {
	printf("TOO FEW SITES\n");
	exit(-1);
      }

    return;
  } /* getnums */


boolean digitchar (int ch) 
{
  return (ch >= '0' && ch <= '9'); 
}


boolean whitechar (int ch)
{ 
  return (ch == ' ' || ch == '\n' || ch == '\t' || ch == '\r'); /* PC-LINEBREAK*/
}


void uppercase (int *chptr)
    /* convert character to upper case -- either ASCII or EBCDIC */
  { /* uppercase */
    int  ch;

    ch = *chptr;
    if ((ch >= 'a' && ch <= 'i') || (ch >= 'j' && ch <= 'r')
                                 || (ch >= 's' && ch <= 'z'))
      *chptr = ch + 'A' - 'a';
  } /* uppercase */


int findch (int c)
  { /* findch */
    int ch;

    while ((ch = getc(INFILE)) != EOF && ch != c) ;
    return  ch;
  } /* findch */

void getyspace (rawdata *rdta)
  { /* getyspace */
    long   size;
    int    i;
    char *y0;

    if (! (rdta->y = (char **) malloc((rdta->numsp + 1) * sizeof(char *)))) 
      {
	printf("ERROR: Unable to obtain space for data array pointers\n");
	exit(-1);
      }

    size = 4 * (rdta->sites / 4 + 1);
    if (! (y0 = (char *) malloc((rdta->numsp + 1) * size * sizeof(char)))) 
      {
	printf("ERROR: Unable to obtain space for data array\n");
	exit(-1);
      }

    rdta->y0 = y0;

    for (i = 0; i <= rdta->numsp; i++) 
      {
	rdta->y[i] = y0;
	y0 += size;
      }

    return;
  } /* getyspace */


void freeyspace(rawdata *rdta)
{
  long   size;
  int    i;
  char *y0;
  for (i = 0; i <= rdta->numsp; i++) 
    {
      free(rdta->y[i]);    
    }
}



boolean setupTree (tree *tr, int nsites, analdef *adef)
  { /* setupTree */
    nodeptr  p0, p, q;
    int      i, j, tips, inter, code;

    if(!adef->useMultipleModel)
      tr->NumberOfModels = 1;

    if(adef->grouping)
      tr->grouped = TRUE;
    else
      tr->grouped = FALSE;

    if(adef->constraint)
      tr->constrained = TRUE;
    else
      tr->constrained = FALSE;
    
    tr->treeID = 0;

    tips  = tr->mxtips;
    inter = tr->mxtips - 1;

    tr->gtrTip       = (likelivector *)malloc(tr->NumberOfModels * 16 * sizeof(likelivector));

    tr->frequencies  = (double *)malloc(tr->NumberOfModels * 4 * sizeof(double));
    tr->EI           = (double *)malloc(tr->NumberOfModels * 12 * sizeof(double));
    tr->EV           = (double *)malloc(tr->NumberOfModels * 16 * sizeof(double));
    tr->EIGN         = (double *)malloc(tr->NumberOfModels * 3 * sizeof(double));
    tr->initialRates = (double *)malloc(tr->NumberOfModels * 5 * sizeof(double));
    tr->gammaRates   = (double *)malloc(tr->NumberOfModels * 4 * sizeof(double));
    tr->alphas       = (double *)malloc(tr->NumberOfModels * sizeof(double));
    tr->fracchanges  = (double *)malloc(tr->NumberOfModels * sizeof(double));
    tr->likelihoods  = (double *)malloc(adef->multipleRuns * sizeof(double));
    tr->treeStringLength = tr->mxtips * (nmlngth+128) + 256 + tr->mxtips * 2;
    tr->tree_string  = (char   *)malloc(tr->treeStringLength * sizeof(char));
    /*TODO, must that be so long ?*/
  
#ifdef PARALLEL_2
    tr->rDesc = (rearrangementDescriptor *)malloc(sizeof(rearrangementDescriptor));
#endif
   

    tr->nameList = (char **)malloc(sizeof(char *) * (tips + 1));    
         

#ifdef TRO
      tr->rearrOrder = (nodeptr *)malloc(3 * tips * sizeof(nodeptr));
#endif      

    if (!(p0 = (nodeptr) malloc((tips + 3*inter) * sizeof(node)))) {
      printf("ERROR: Unable to obtain sufficient tree memory\n");
      return  FALSE;
      }


    if (!(tr->nodep = (nodeptr *) malloc((2*tr->mxtips) * sizeof(nodeptr)))) {
      printf("ERROR: Unable to obtain sufficient tree memory, too\n");
      return  FALSE;
      }

    
    tr->nodep[0] = (node *) NULL;    /* Use as 1-based array */

    for (i = 1; i <= tips; i++) {    /* Set-up tips */
      p = p0++;
      p->x      =  NULL;
      p->tip    = (char *) NULL;      
      p->number = i;
      p->next   = p;
      p->back   = (node *) NULL;      
      tr->nodep[i] = p;
      }

    for (i = tips + 1; i <= tips + inter; i++) { /* Internal nodes */
      q = (node *) NULL;
      for (j = 1; j <= 3; j++) {
        p = p0++;
        p->x      =  NULL;
        p->tip    = (char *) NULL;
        p->number = i;

        p->next   = q;
        p->back   = (node *) NULL;
        q = p;
        }
      p->next->next->next = p;
      tr->nodep[i] = p;
      }

    tr->likelihood  = unlikely;
    tr->start       = (node *) NULL;
    tr->ntips       = 0;
    tr->nextnode    = 0;
    tr->prelabeled  = TRUE;
    tr->smoothed    = FALSE;
   
    return TRUE;
  } /* setupTree */

void freeTreeNode (nodeptr p)   /* Free tree node (sector) associated data */
  { /* freeTreeNode */
    if (p) {
      if (p->x) 
	{
	  free(p->x);      
        }
      }
  } /* freeTree */

void freeTree (tree *tr)
  { /* freeTree */
    nodeptr  p, q;
    int  i, tips, inter;

    tips  = tr->mxtips;
    inter = tr->mxtips - 1;

    for (i = 1; i <= tips; i++) freeTreeNode(tr->nodep[i]);

    for (i = tips + 1; i <= tips + inter; i++) {
      if (p = tr->nodep[i]) {
        if (q = p->next) {
          freeTreeNode(q->next);
          freeTreeNode(q);
          }
        freeTreeNode(p);
        }
      }
    

    free(tr->nodep[1]);       /* Free the actual nodes */
  } /* freeTree */



boolean getdata (boolean reRead, analdef *adef, rawdata *rdta, tree *tr)
{
  int   i, j, k, l, basesread, basesnew, ch, my_i;
  int   meaning[256];
  char *nameptr;
  boolean  allread, firstpass;
  char buffer[300]; 
  int len;
   
  for (i = 0; i <= 255; i++) 
    meaning[i] = -1;

  meaning['A'] =  1;
  meaning['B'] = 14;
  meaning['C'] =  2;
  meaning['D'] = 13;
  meaning['G'] =  4;
  meaning['H'] = 11;
  meaning['K'] = 12;
  meaning['M'] =  3;
  meaning['N'] = 15;
  meaning['O'] = 15;
  meaning['R'] =  5;
  meaning['S'] =  6;
  meaning['T'] =  8;
  meaning['U'] =  8;
  meaning['V'] =  7;
  meaning['W'] =  9;
  meaning['X'] = 15;
  meaning['Y'] = 10;
  meaning['?'] = 15;
  meaning['-'] = 15;	
  
  basesread = basesnew = 0;

  allread = FALSE;
  firstpass = TRUE;
  ch = ' ';

  while (! allread) 
    {
      for (i = 1; i <= tr->mxtips; i++) 
	{   	  
	  if (firstpass) 
	    {                      	       
	      ch = getc(INFILE);
	      while(ch == ' ' || ch == '\n' || ch == '\t' || ch == '\r') /* PC-LINEBREAK*/
		{
		  ch = getc(INFILE);		  
		}	      
	      my_i = 0;	      

	      do 
		{
		  buffer[my_i] = ch;		  
		  ch = getc(INFILE);		   
		  my_i++;
		  if(my_i >= nmlngth)
		    {
		      printf("ERROR Names are to long, adapt constant nmlngth in\n");
		      printf("axml.h AND dnapars.c, current setting %d\n", nmlngth);
		      exit(-1);
		    }		 
		}
	      while(ch !=  ' ' && ch != '\n' && ch != '\t' && ch != '\r'); /* PC-LINEBREAK*/
		
	      if(!reRead)
		{
		  buffer[my_i] = '\0';
			
		  len = strlen(buffer) + 1;
			       
		  tr->nameList[i] = (char *)malloc(sizeof(char) * len);

		  strcpy(tr->nameList[i], buffer);		
		}
	    }

	  j = basesread;
	  while ((j < rdta->sites) && ((ch = getc(INFILE)) != EOF) && (ch != '\n') && (ch != '\r')) /* PC-LINEBREAK*/
	    {
	      uppercase(& ch);
	      if ((meaning[ch] != -1) || ch == '.') 
		{
		  j++;
		  if (ch == '.') 
		    {
		      if (i != 1) ch = rdta->y[1][j];
		      else 
			{
			  printf("ERROR: Dot (.) found at site %d of sequence 1\n", j + 1);
			  return  FALSE;
			}
		    }
		  rdta->y[i][j] = ch;
		}
	      else 
		if(whitechar(ch) || digitchar(ch)) ;
		else 
		  {
		    printf("ERROR: Bad base (%c) at site %d of sequence %d\n",
			   ch, j + 1, i);
		    return  FALSE;
		  }
	    }

	    if (ch == EOF) 
	      {
		printf("ERROR: End-of-file at site %d of sequence %d\n", j + 1, i);
		return  FALSE;
	      }

	    if (! firstpass && (j == basesread)) i--; 
	    else 
	      if (i == 1) basesnew = j;
	      else 
		if (j != basesnew) 
		  {
		    printf("ERROR: Sequences out of alignment\n");		    
		    printf("%d (instead of %d) residues read in sequence %d %s\n",
			   j - basesread, basesnew - basesread, i, tr->nameList[i]);
		    return  FALSE;
		  }
	    while (ch != '\n' && ch != EOF && ch != '\r') ch = getc(INFILE);  /* flush line *//* PC-LINEBREAK*/
	}                                                  /* next sequence */
      firstpass = FALSE;
      basesread = basesnew;
      allread = (basesread >= rdta->sites);
    }

 
    
  for (j = 1; j <= tr->mxtips; j++)    
    for (i = 1; i <= rdta->sites; i++) 	
      rdta->y[j][i] = meaning[rdta->y[j][i]];		
	    	
    return  TRUE;
  } /* getdata */



void inputweights (analdef *adef, rawdata *rdta, cruncheddata *cdta)
    /* input the character weights 0, 1, 2 ... 9, A, B, ... Y, Z */
  { /* inputweights */
    int i, ch, w, fres;
    FILE *weightFile;
    int *wv = (int *)malloc(sizeof(int) *  rdta->sites + 1);
  

    weightFile = fopen(weightFileName, "r");
    if (!weightFile)
      {
	printf( "Could not open weight file: %s\n", weightFileName);
	exit(1);
      }
     
    i = 1;
    
    while((fres = fscanf(weightFile,"%d", &w)) != EOF)
      {
	if(!fres)
	  {
	    printf("error reading weight file probably encountered a non-integer weight value\n");
	    exit(-1);
	  }
	wv[i] = w;
	i++;	
      }
    
   
    if(i != (rdta->sites + 1))
      {
	printf("number %d of weights not equal to number %d of alignment columns\n", i, rdta->sites);
	exit(-1);
      }

    cdta->wgtsum = 0;
    for(i = 1; i <= rdta->sites; i++) 
      {     
	rdta->wgt[i] = wv[i];
	cdta->wgtsum += rdta->wgt[i];
      }

    fclose(weightFile);
    free(wv);
  } /* inputweights */



void getinput (boolean reRead, analdef *adef, rawdata *rdta, cruncheddata *cdta, tree *tr)
{ 
  int i,j;
  getnums(rdta);

  if(!reRead)
    {          
      /* TODO put some conditionals here depending on input params */

      tr->mxtips         = rdta->numsp;
      rdta->wgt          = (int *)    malloc((rdta->sites + 1) * sizeof(int));
      rdta->wgt2         = (int *)    malloc((rdta->sites + 1) * sizeof(int));
      cdta->alias        = (int *)    malloc((rdta->sites + 1) * sizeof(int));
      cdta->reAlias      = (int *)    malloc((rdta->sites + 1) * sizeof(int));
      cdta->aliaswgt     = (int *)    malloc((rdta->sites + 1) * sizeof(int));      
      cdta->rateCategory = (int *)    malloc((rdta->sites + 1) * sizeof(int)); 
      tr->model          = (int *)    malloc((rdta->sites + 1) * sizeof(int));
      tr->saveModel      = (int *)    malloc((rdta->sites + 1) * sizeof(int));
      cdta->wr           = (double *) malloc((rdta->sites + 1) * sizeof(double));  
      cdta->wr2          = (double *) malloc((rdta->sites + 1) * sizeof(double));  	
      cdta->patrat       = (double *) malloc((rdta->sites + 1) * sizeof(double));
      cdta->patratStored = (double *) malloc((rdta->sites + 1) * sizeof(double));             
    }
      
  if(!adef->useWeightFile)
    {
      for (i = 1; i <= rdta->sites; i++) 
	rdta->wgt[i] = 1;
     
      cdta->wgtsum = rdta->sites;   
    }
  else    
    inputweights(adef, rdta, cdta);
    
  
  if(adef->useMultipleModel)        
    parsePartitions(adef, rdta, cdta, tr, reRead);    

  if(!reRead)
    {
      getyspace(rdta);
      setupTree(tr, rdta->sites, adef);
    }
  
  getdata(reRead, adef, rdta, tr);
        
  return;
} 

void makeboot (analdef *adef, rawdata *rdta, cruncheddata *cdta)
  {
    int  i, j, nonzero;
#ifdef PARALLEL
    long seed;
#endif
   
    nonzero = 0;
   
    for (i = 1; i <= rdta->sites; i++)  
      {
	if (rdta->wgt[i] > 0) nonzero++;		
      }

    for (j = 1; j <= nonzero; j++) 
      cdta->aliaswgt[j] = 0;     

#ifdef PARALLEL 
    seed = (long) gettimeSrand();
    for (j = 1; j <= nonzero; j++)
      cdta->aliaswgt[(int) (nonzero*randum(& seed)) + 1]++;
#endif    

#ifndef PARALLEL
    for (j = 1; j <= nonzero; j++)
      cdta->aliaswgt[(int) (nonzero*randum(& adef->boot)) + 1]++;
    
    /*
      Code for testing bootstrap. do not remove !
      srand(1234);
      for (j = 1; j <= nonzero; j++)
      cdta->aliaswgt[(rand() % nonzero) + 1]++;     
    */
#endif

    j = 0;
    cdta->wgtsum = 0;
    for (i = 1; i <= rdta->sites; i++) 
      {
	if (rdta->wgt[i] > 0)
	  cdta->wgtsum += (rdta->wgt2[i] = rdta->wgt[i] * cdta->aliaswgt[++j]);
	else
	  rdta->wgt2[i] = 0;
      }        
  } 




void sitesort (rawdata *rdta, cruncheddata *cdta, tree *tr, analdef *adef)   
  { 
    int  gap, i, j, jj, jg, k, n, nsp;
    int  *index, *category;
    boolean  flip, tied;
    char  **data;

    
    if(adef->useMultipleModel)
      category = tr->model;

    index    = cdta->alias;
    data     = rdta->y;
    n        = rdta->sites;
    nsp      = rdta->numsp;
    index[0] = -1;
    for (gap = n / 2; gap > 0; gap /= 2) 
      {
	for (i = gap + 1; i <= n; i++) 
	  {
	    j = i - gap;

	    do 
	      {
		jj = index[j];
		jg = index[j+gap];
		if(adef->useMultipleModel)
		  {		   
		     flip = (category[jj] >  category[jg]);
		     tied = (category[jj] == category[jg]);		    
		  }
		else
		  {		    
		    flip = 0;
		    tied = 1;
		  }
	
		for (k = 1; (k <= nsp) && tied; k++) 
		  {
		    flip = (data[k][jj] >  data[k][jg]);
		    tied = (data[k][jj] == data[k][jg]);
		  }

		if (flip) 
		  {
		    index[j]     = jg;
		    index[j+gap] = jj;
		    j -= gap;
		  }
	      } 
	    while (flip && (j > 0));

        }  
      }       
  } 


void sitecombcrunch (rawdata *rdta, cruncheddata *cdta, tree *tr, analdef *adef)
    
  { 
    int  i, sitei, j, sitej, k;
    int gaps = 0;
    boolean  tied;
    int *aliasModel;

    
    if(adef->useMultipleModel)
      aliasModel = (int*)malloc(sizeof(int) * (rdta->sites + 1));

    i = 0;    
    cdta->alias[0] = cdta->alias[1];
    cdta->aliaswgt[0] = 0;   
    
    
    for (j = 1; j <= rdta->sites; j++) 
      {
	sitei = cdta->alias[i];
	sitej = cdta->alias[j];

	if(adef->useMultipleModel)
	  {
	    tied = (tr->model[sitei] == tr->model[sitej]);
	  }
	else
	  tied = 1;

	for (k = 1; tied && (k <= rdta->numsp); k++)
	  tied = (rdta->y[k][sitei] == rdta->y[k][sitej]);

	if (tied) 
	  {
	    cdta->aliaswgt[i] += rdta->wgt2[sitej];	   
	    if(adef->useMultipleModel)
	      aliasModel[i] = tr->model[sitej];
	       
	  }
	else 
	  {
	    if (cdta->aliaswgt[i] > 0) i++;
	    cdta->aliaswgt[i] = rdta->wgt2[sitej];
	    cdta->alias[i] = sitej;
	    if(adef->useMultipleModel)
	      aliasModel[i] = tr->model[sitej];
	   
	  }
      }

    cdta->endsite = i;
    if (cdta->aliaswgt[i] > 0) cdta->endsite++;       


    if(adef->useMultipleModel)
      {       
	for(i = 0; i <= rdta->sites; i++)
	  tr->model[i] = aliasModel[i];	  
      }


    if(adef->useMultipleModel)
      free(aliasModel);	
  } 




boolean makeweights (analdef *adef, rawdata *rdta, cruncheddata *cdta, tree *tr)
    /* make up weights vector to avoid duplicate computations */
  { /* makeweights */
    int  i;

    if (adef->boot)  
      makeboot(adef, rdta, cdta);
    else 
      {
	for (i = 1; i <= rdta->sites; i++)  
	  rdta->wgt2[i] = rdta->wgt[i];
      }

    for (i = 1; i <= rdta->sites; i++)  
      cdta->alias[i] = i;
        
    sitesort(rdta, cdta, tr, adef);
    sitecombcrunch(rdta, cdta, tr, adef);
      
    return TRUE;
  } /* makeweights */


boolean makevalues (rawdata *rdta, cruncheddata *cdta, tree *tr, analdef *adef)
    /* set up fractional likelihoods at tips */
  { /* makevalues */
    double  temp, wtemp;
    int  i, j, model;
    char buf;      

    for (i = 1; i <= rdta->numsp; i++)       
      for (j = 0; j < cdta->endsite; j++) 	  	   
	rdta->y[i-1][j] = rdta->y[i][cdta->alias[j]];
	                              
    if(adef->useMultipleModel)
      {
	tr->modelIndices = (int **)malloc(tr->NumberOfModels * sizeof(int *));
	
	for(i = 0; i < tr->NumberOfModels; i++)
	  tr->modelIndices[i] = (int *)malloc(2 * sizeof(int));

	tr->modelIndices[0][0] = 0;	

	model = tr->model[0];
	i = 1;

	while(i <  cdta->endsite)
	  {
	    if(tr->model[i] != model)
	      {
		tr->modelIndices[model][1]     = i;
		tr->modelIndices[model + 1][0] = i;
		model = tr->model[i];
	      }
	    i++;
	  }

	tr->modelIndices[tr->NumberOfModels - 1][1] = cdta->endsite;
      }
    else
      {
	tr->modelIndices = (int **)malloc(1 * sizeof(int *));
	
	tr->modelIndices[0] = (int *)malloc(2 * sizeof(int));

	tr->modelIndices[0][0] = 0;
	tr->modelIndices[0][1] = cdta->endsite;
      }
    
    return TRUE;
  } /* makevalues */




void linkdata2tree (rawdata *rdta, cruncheddata *cdta, tree *tr)    
{ 
  int  i;
  
  tr->rdta       = rdta;
  tr->cdta       = cdta;  
    
  for (i = 1; i <= tr->mxtips; i++)   
    tr->nodep[i]->tip = &(rdta->y[i-1][0]);      
}


void allocNodex (tree *tr, analdef *adef)
{
  nodeptr  p;
  int  i;
    
  for (i = tr->mxtips + 1; (i <= 2*(tr->mxtips) - 2); i++) 
    {
      p = tr->nodep[i];
      allocCount++;
      if(adef->model == M_GTRGAMMA)
	{	       
#ifndef SEVS
	  p->x = malloc(tr->cdta->endsite * sizeof(gammalikelivector));	    	  
#endif
#ifdef SEVS
	  p->x = malloc((tr->cdta->endsite + 16) * sizeof(gammalikelivector));	
#endif
	}
      else
	{      	
	  p->x = malloc(tr->cdta->endsite * sizeof(likelivector));	    		  
	}		
    }  
} 


void freeNodex(tree *tr)
{
  nodeptr  p;
  int  i;  
  
  for (i = tr->mxtips + 1; (i <= 2*(tr->mxtips) - 2); i++) 
    {
      p = tr->nodep[i];
      while(!p->x)
	p = p->next;
      allocCount--;
      free(p->x);
      p->x = NULL;
    }

  if(allocCount != 0)
    {
      printf("FATAL error\n");
      exit(-1);
    }

}

void initAdef(analdef *adef)
{
  adef->model = M_GTRCAT;  
  adef->max_rearrange = 21;  
  adef->stepwidth = 5;
  adef->initial = adef->bestTrav = 10;
  adef->initialSet = FALSE;
  adef->restart = FALSE;
  adef->mode = BIG_RAPID_MODE;
  adef->categories = 25; 
  adef->boot    =           0;
  adef->useWeightFile = FALSE;
  adef->checkpoints = 0;
  adef->startingTreeOnly = 0;
  adef->rapidParsimony = 0;
  adef->useMixedModel = 0;
  adef->multipleRuns = 1;
  adef->useMultipleModel = FALSE;
  adef->likelihoodEpsilon = 0.1;
  adef->constraint        = FALSE;
  adef->grouping =          FALSE; 
  adef->randomStartingTree = FALSE;
  adef->parsimonySeed = 0;
  /* RELL */
  adef->THOROUGH = 0;
  adef->sdThreshold = -1.0;
  adef->lhThreshold = 14.5;
  /* RELL end */
}

void get_args(int argc, char *argv[], boolean print_usage, analdef *adef, rawdata *rdta)
{ /* get_args */
  int        c;
  boolean    bad_opt=FALSE;
  int        i, time, temperature;
  char       buf[2048];
  char       model[2048] = "";
  char       modelChar;
  double likelihoodEpsilon;
  int nameSet = 0, 
    alignmentSet = 0, 
    catSet = 0,
    checkpointSet = 0,
    startingTreeOnly = 0,
    multipleRuns = 0,
    parsimonySeed = 0,
    constraintSet = 0,
    treeSet = 0,
    groupSet = 0;
  
  
  run_id[0] = 0;
  workdir[0] = 0;
  seq_file[0] = 0;
  tree_file[0] = 0;
  model[0] = 0;
  weightFileName[0] = 0;
  modelFileName[0] = 0;
 
  while(!bad_opt && ((c=getopt(argc,argv,"z:l:o:g:r:e:a:b:c:f:i:m:t:w:s:n:q:#:p:vdyjhk"))!=-1) )
    {
    switch(c) 
      {
      case 'z':	
	strcpy(bootStrapFile, optarg);
	break;
      case 'l':
	sscanf(optarg,"%lf", &(adef->lhThreshold));       
	adef->sdThreshold = -1.0;
	break;
      case 'o':
	sscanf(optarg,"%lf", &(adef->sdThreshold));      
	adef->lhThreshold = -1.0;
	break;
      case 'k':
	adef->THOROUGH = 1;
	break;
      case 'd':
	adef->randomStartingTree = TRUE;
	break;
      case 'g':
	strcpy(tree_file, optarg);
	adef->grouping = TRUE;
	adef->restart  = TRUE;      
	groupSet = 1;
	break;
      case 'r':	
	strcpy(tree_file, optarg);
	adef->restart = TRUE;
	adef->constraint = TRUE;
	constraintSet = 1;
	break;
      case 'e':      
	sscanf(optarg,"%lf", &likelihoodEpsilon);
	adef->likelihoodEpsilon = likelihoodEpsilon;      
	break;
      case 'q':
	strcpy(modelFileName,optarg);
	adef->useMultipleModel = TRUE;      
        break;      
      case 'p':
	sscanf(optarg,"%d", &parsimonySeed);
	adef->parsimonySeed = parsimonySeed;
	break;
      case '#':
	sscanf(optarg,"%d", &multipleRuns);
	adef->multipleRuns = multipleRuns;
	break;
      case 'v':
	printVersionInfo();
	exit(1);
      case 'y':
	adef->startingTreeOnly = 1;
	break;
      case 'h':
	printREADME();
	exit(1);
      case 'j':	
	adef->checkpoints = 1;
	break;
      case 'a':
	strcpy(weightFileName,optarg);
	adef->useWeightFile = TRUE;
        break;            
      case 'b':
	sscanf(optarg,"%ld", &adef->boot);       
	break;
      case 'c':
	sscanf(optarg, "%d", &adef->categories);
	catSet = 1;
	break;	      	
      case 'f': 
	sscanf(optarg, "%c", &modelChar);
	switch(modelChar)
	  {
	  case 'b': adef->mode = CALC_BIPARTITIONS; break;
	  case 'e': adef->mode = TREE_EVALUATION; break;	  
	  case 'd': adef->mode = BIG_RAPID_MODE; break;	 
	  case 'r': adef->mode = RELL_BOOTSTRAP; break;	   
	  default: 
	    {
	      printf("Error select on of the following algorithms:\n");	     	     
	      printf("-f e : optimization of model parameters and branch lengths\n");	      	     
	      printf("-f d : very fast hill-climbing for huge data\n");
	      /*
	      printf("-f b : computation of bipartitions\n");	      	     
	      printf("-f d : fast RELL bootstrap\n");
	      */
	      exit(-1);
	    }
	  }
	break;      
      case 'i':
	sscanf(optarg, "%d", &adef->initial);
	adef->initialSet = TRUE;
	break;     
      case 'n':
        strcpy(run_id,optarg);
	nameSet = 1;
        break;
      case 'w':
        strcpy(workdir,optarg);
        break;                 
      case 't':
	strcpy(tree_file, optarg);
	adef->restart = TRUE;
	treeSet = 1;
	break;
      case 's':
	strcpy(seq_file, optarg);
	alignmentSet = 1;
	break;
      case 'm':
	strcpy(model,optarg);
	if(strcmp(model, "GTRGAMMA\0") == 0)
	  {
	    adef->model = M_GTRGAMMA;
	    goto l_1;
	  }
	if(strcmp(model, "GTRCAT\0") == 0)
	  {
	    adef->model = M_GTRCAT;
	    goto l_1;
	  }
	if(strcmp(model, "GTRMIX\0") == 0)
	  {
	    adef->model = M_GTRCAT;
	    adef->useMixedModel = 1;
	    goto l_1;
	  }
	printf("Model %s does not exist\n", model);
	printf("please use either GTRCAT or GTRGAMMA or GTRMIX for DNA DATA\n");
	exit(-1);
      l_1:
	break;      
      default:
        bad_opt = TRUE;
        break;
    }    
  }     

  if(adef->mode ==  RELL_BOOTSTRAP && (!adef->restart))
    {
      printf("\n Error the RELL bootstrap algorithm needs a set of starting trees with -t\n");  
      exit(-1);
    }

  if(adef->restart && adef->randomStartingTree)
    {
      printf("\n Error you specified a starting tree with -t AND the computation\n");
      printf("of a random starting tree	with -d for the same run\n");
      exit(-1);
    }

  if(treeSet && constraintSet)
    {
      printf("\n Error you specified a binary constraint tree AND a starting tree for the same run\n");
      exit(-1);
    }


  if(treeSet && groupSet)
    {
      printf("\n Error you specified a multifurcating constraint tree AND a starting tree for the same run\n");
      exit(-1);
    }


  if(groupSet && constraintSet)
    {
      printf("\n Error you specified a bifurcating constraint tree AND a multifurcating constraint tree for the same run\n");
      exit(-1);
    }

  if(adef->restart && adef->startingTreeOnly)
    {
      printf("\n Error conflicting options: you want to compute only a parsimony starting tree with -y\n");
      printf(" while you actually specified a starting tree with -t %s\n", tree_file);
      exit(-1);
    }
  
  if(adef->mode == TREE_EVALUATION && (!adef->restart))
    {
      printf("\n Error: please specify a treefile for the tree you want to evaluate with -t\n");
      exit(-1);
    }

#ifdef PARALLEL
  if(adef->mode == TREE_EVALUATION)
    {
      printf("Tree Evaluation mode (-f e) noot implemented for the MPI-Version\n");
      exit(-1);
    }

  if(adef->mode == RELL_BOOTSTRAP)
    {
      printf("RELL Bootstrap mode (-f e) not implemented for the MPI-Version\n");
      exit(-1);
    }

   if(adef->mode == CALC_BIPARTITIONS)
     {
       printf("Computation of bipartitions (-f b) not implemented for the MPI-Version\n");
       exit(-1);
     }
   
   if(adef->multipleRuns == 1)
     {
       printf("WARNING: you are running the parallel MPI program but only want to compute one tree\n");
       printf("Typically you should specify a number of trees greater than 1 with the -# option\n");
     }
#endif

   if(adef->boot && adef->model == M_GTRCAT && adef->useMixedModel)
     {
       printf("WARNING: you want to use GTRMIX for a boootstrap run\n");
       printf("this will be ignored by RAxML since it assumes that you are\n");
       printf("only interested in the bootstrap trees, not their likelihoods\n");
     }
   
   if(adef->mode == TREE_EVALUATION && adef->model == M_GTRCAT)
     {
       printf("\n Error: No tree evaluation with GTRCAT possible\n");
       printf("the GTRCAT likelihood values are instable at present and should not\n");
       printf("be used to compare trees based on ML values\n");
       exit(-1);
     }

  if(!nameSet)
    {
      printf("\n Error: please specify a name for this run with -n\n");
      exit(-1);
    }
    
  if(! alignmentSet)
    {
      printf("\n Error: please specify an alignment for this run with -s\n");
      exit(-1);
    }

  if(workdir[0]==0 || workdir[0] != '/') 
    {
      getcwd(buf,sizeof(buf));
      if( buf[strlen(buf)-1] != '/') strcat(buf,"/");
      strcat(buf,workdir);
      if( buf[strlen(buf)-1] != '/') strcat(buf,"/");
      strcpy(workdir,buf);
    }

  if(bad_opt) 
    {      
      exit(-1);
    }
    return;
}


void makeFileNames(tree *tr, analdef *adef, int argc, char *argv[])
{

  strcpy(permFileName,       workdir);    
  strcpy(resultFileName,     workdir);
  strcpy(logFileName,        workdir);
  strcpy(checkpointFileName, workdir);
  strcpy(infoFileName,       workdir);
  strcpy(randomFileName,     workdir);
  strcpy(rellInfoFileName,   workdir);
  strcpy(rellBootFileName,   workdir);
  strcpy(bootstrapFileName,  workdir);

  strcat(permFileName,       "RAxML_parsimonyTree.");
  strcat(resultFileName,     "RAxML_result.");
  strcat(logFileName,        "RAxML_log.");
  strcat(checkpointFileName, "RAxML_checkpoint.");
  strcat(infoFileName,       "RAxML_info.");
  strcat(randomFileName,     "RAxML_randomTree.");
  strcat(rellInfoFileName,   "RAxML_RELL_INFO.");
  strcat(rellBootFileName,   "RAxML_RELL.");
  strcat(bootstrapFileName,  "RAxML_bootstrap.");  

  strcat(permFileName,       run_id);
  strcat(resultFileName,     run_id);
  strcat(logFileName,        run_id);
  strcat(checkpointFileName, run_id);
  strcat(infoFileName,       run_id);    
  strcat(randomFileName,     run_id);  
  strcat(rellInfoFileName,   run_id);
  strcat(rellBootFileName,   run_id);
  strcat(bootstrapFileName,  run_id); 

  if(filexists(infoFileName))
    {
      printf("RAxML output files with the run ID <%s> already exist \n", run_id);
      printf("in directory %s ...... exiting\n", workdir);
#ifdef PARALLEL 
      MPI_Finalize();
      exit(-1);
#else
      exit(-1);
#endif
    } 
}


void readData(boolean reRead, analdef *adef, rawdata *rdta, cruncheddata *cdta, tree *tr)
{
  INFILE = fopen(seq_file, "r");
  
  if (!INFILE)
    {
      printf( "Could not open sequence file: %s\n", seq_file);
      exit(-1);
    }   
  getinput(reRead, adef, rdta, cdta, tr); 
  
  fclose(INFILE);   
}



/***********************reading and initializing input ******************/


/********************PRINTING various INFO **************************************/

void printVersionInfo()
{
  printf("\nThis is %s version %s released by Alexandros Stamatakis in %s\n\n",  programName, programVersion, programDate);
}


/* TODO: update */

void printREADME()
{
  printVersionInfo();  
  printf("\n");
  printf("Please also consult the RAxML-manual\n");
  printf("To report bugs send an email to stamatak@ics.forth.gr\n\n\n");
  
  printf("raxmlHPC[-MPI|-OMP] [-a weightFileName] [-b bootstrapRandomNumberSeed] [-c numberOfCategories]\n");
  printf("                    [-d] [-e likelihoodEpsilon] [-f d|e] [-g groupingFileName]\n");
  printf("                    [-h] [-i initialRearrangementSetting] [-j] [-m GTRCAT|GTRMIX|GTRGAMMA]\n");
  printf("                    [-q multipleModelFileName] [-t userStartingTree] [-w workingDirectory]\n");
  printf("                    [-v] [-y] [-# numberOfRuns]\n");
  printf("                    -s sequenceFileName -n outputFileName\n");
  printf("\n");
  printf("	-a 	Specify a column weight file name to assign individual weights to each column of \n");
  printf("		the alignment. Those weights must be integers separated by any type and number \n");
  printf("		of whitespaces whithin a separate file, see file \"example_weights\" for an example.\n");
  printf("\n");
  printf("	-b	Specify an integer number (random seed) and turn on bootstrapping\n");
  printf("\n");
  printf("		DEFAULT: OFF\n");
  printf("\n");
  printf("	-c 	Specify number of distinct rate catgories for RAxML when modelOfEvolution\n");
  printf("		is set to GTRCAT or GTRMIX\n");
  printf("		Individual per-site rates are categorized into numberOfCategories rate \n");
  printf("		categories to accelerate computations. \n");
  printf("\n");
  printf("		DEFAULT: 25\n");
  printf("\n");
  printf("	-d	start ML optimization from random starting tree \n");
  printf("\n");
  printf("		DEFAULT: OFF\n");
  printf("\n");
  printf("	-e	set model optimization precision in log likelihood units for final\n"); 
  printf("		optimization of tree topology under GTRMIX or GTRGAMMA\n");
  printf("\n");
  printf("		DEFAULT: 0.1\n");
  printf("\n");
  printf("	-f	select algorithm:\n");
  printf("\n");
  printf("		\"-f d\": rapid hill-climbing \n");
  printf("		\"-f e\": optimize model+branch lengths for given input tree under GTRGAMMA only\n"); 
  printf("\n");
  printf("		DEFAULT: rapid hill climbing\n");  
  printf("\n");
  printf("	-g	specify the file name of a multifurcating constraint tree\n");
  printf("		this tree does not need to be comprehensive, i.e. must not contain all taxa\n");
  printf("\n");
  printf("	-i  	Initial rearrangement setting for the subsequent application of topological \n");
  printf("		changes phase\n");
  printf("\n");
  printf("		DEFAULT: determined by program\n");
  printf("\n");
  printf("	-j	Specifies if checkpoints will be written by the program. If checkpoints \n");
  printf("		(intermediate tree topologies) shall be written by the program specify \"-j\"\n");
  printf("\n");
  printf("		DEFAULT: OFF\n");
  printf("\n");
  printf("	-h	Display this help message.\n");  
  printf("\n");
  printf("	-m      Model of Nucleotide Substitution: \n");  
  printf("\n");
  printf("		\"-m GTRCAT\":   GTR + Optimization of substitution rates +  Optimization of site-specific\n");
  printf("                               evolutionary rates which are categorized into numberOfCategories distinct \n");
  printf("                               rate categories for greater computational efficiency\n");
  printf("                               if you do a multiple analysis with  \"-#\" but without bootstrapping the program\n");
  printf("                               will use GTRMIX instead\n");
  printf("		\"-m GTRGAMMA\": GTR + Optimization of substitution rates + GAMMA model of rate \n");
  printf("                               heterogeneity (alpha parameter will be estimated)\n");
  printf("		\"-m GTRMIX\":   Inference of the tree under GTRCAT\n");
  printf("                               and thereafter evaluation of the final tree topology under GTRGAMMA\n");
  printf("\n");
  printf("		DEFAULT: GTRCAT\n");
  printf("\n");
  printf("	-n	Specifies the name of the output file.	\n"); 
  printf("\n"); 
  printf("	-q	Specify the file name which contains the assignment of models to alignment\n"); 
  printf("		partitions for multiple models of substitution. For the syntax of this file\n");
  printf("		please consult the manual.\n"); 
  printf("\n"); 
  printf("	-r	specify the file name of a binary constraint tree.\n");
  printf("		this tree does not need to be comprehensive, i.e. must not contain all taxa\n");
  printf("\n");  
  printf("	-s 	specify the name of the alignment data file in PHYLIP format\n");
  printf("\n");
  printf("	-t      Specify a user starting tree file name in Newick format\n");
  printf("\n");
  printf("	-v      Display version information\n");  
  printf("\n");
  printf("	-w 	Name of the working directory where RAxML will write its output files\n");
  printf("\n");
  printf("		DEFAULT: current directory\n"); 
  printf("\n");
  printf("	-y	If you want to only compute a parsimony starting tree with RAxML specify \"-y\",\n");
  printf("		the program will exit after computation of the starting tree\n");
  printf("\n");
  printf("		DEFAULT: OFF\n");
  printf("\n");  
  printf("	-#	Specify the number of alternative runs on distinct starting trees\n");
  printf("		In combination with the \"-b\" option, this will invoke a multiple boostrap analysis\n");
  printf("\n");
  printf("		DEFAULT: 1 single analysis\n");
  printf("\n\n\n\n");

}

void printModelAndProgramInfo(tree *tr, analdef *adef, int argc, char *argv[])
{ 
  if(processID == 0)
    {

      int i;
      
      FILE *infoFile = fopen(infoFileName, "a");  
      
      switch(adef->mode)
	{       
	case TREE_EVALUATION : 
	  printf("RAxML Model Optimization up to an accuracy of %f log likelihood units\n\n", adef->likelihoodEpsilon);
	  fprintf(infoFile, "RAxML Model Optimization up to an accuracy of %f log likelihood units\n\n", adef->likelihoodEpsilon);
	  break;     
	case  BIG_RAPID_MODE:
	  printf("RAxML rapid hill-climbing mode for huge data\n\n"); 
	  fprintf(infoFile, "RAxML   rapid hill-climbing mode for huge data\n\n");
	  break;
	case RELL_BOOTSTRAP:
	  printf("TODO RELL-BOOTSTRAP\n");
	  printf("RELL parameters: Rells per Tree %d, sd cutoff %f lh cutoff %f Thourough %s MaxTrav %d\n", RELLs_PER_TREE,  adef->sdThreshold,  
		 adef->lhThreshold, (adef->THOROUGH)?"ON":"OFF", adef->initial);
	  fprintf(infoFile, "RELL parameters: Rells per Tree %d, sd cutoff %f lh cutoff %f Thourough %s MaxTrav %d\n", RELLs_PER_TREE,  adef->sdThreshold,  
		  adef->lhThreshold, (adef->THOROUGH)?"ON":"OFF", adef->initial);
	  break;
	case CALC_BIPARTITIONS:
	  printf("TODO BIPARTITIONS-COMPUTATION\n");
	  break;
	}
      
      if(tr->NumberOfModels > 1)
	{
	  int model;
	  
	  printf("Multiple Model Mode: Using %d distinct models\n", tr->NumberOfModels);
	  fprintf(infoFile, "Multiple Model Mode: Using %d distinct models\n", tr->NumberOfModels);
#ifdef DEBUG_MODEL
	  for(model = 0; model < tr->NumberOfModels; model++)
	    {
	      printf(           "Partition %d: position after sorting %d to %d\n", model, tr->modelIndices[model][0], tr->modelIndices[model][1]);
	      fprintf(infoFile, "Partition %d: position after sorting %d to %d\n", model, tr->modelIndices[model][0], tr->modelIndices[model][1]); 
	    }
#endif
	  
	  printf(           "\n\n");
	  fprintf(infoFile, "\n\n");
	}
      
      if(adef->boot)
	{
	  if(adef->multipleRuns > 1)
	    {
	      printf("Executing %d non-parametric bootstrap inferences\n\n", adef->multipleRuns);
	      fprintf(infoFile, "Executing %d non-parametric bootstrap inferences\n\n", adef->multipleRuns);
	    }
	  else
	    {
	      printf("Executing %d non-parametric bootstrap inference\n\n", adef->multipleRuns);
	      fprintf(infoFile, "Executing %d non-parametric bootstrap inference\n\n", adef->multipleRuns);
	    }
	}
      else
	{
	  char treeType[1024];

	  if(adef->restart)
	    strcpy(treeType, "user-specifed");
	  else
	    {
	      if(adef->randomStartingTree)		
		strcpy(treeType, "distinct complete random");
	      else
		strcpy(treeType, "distinct randomized MP");
	    }


	  if(adef->multipleRuns > 1)
	    {
	      printf("Executing %d inferences on the originial alignment using %d %s trees\n\n", adef->multipleRuns, adef->multipleRuns, treeType);
	      fprintf(infoFile, "Executing %d inferences on the originial alignment using %d %s trees\n\n", adef->multipleRuns, adef->multipleRuns, treeType);
	    }
	  else
	    {
	      printf("Executing %d inference on the originial alignment using a %s tree\n\n", adef->multipleRuns, treeType);
	      fprintf(infoFile, "Executing %d inference on the originial alignment using a %s tree\n\n", adef->multipleRuns, treeType);
	    }
	}
      
      
      switch(adef->model)
	{    
	case M_GTRCAT:
	  if(adef->useMixedModel)
	    { 
	      printf( "GTR mixed model of nucleotide substitution\n"); 
	      printf( "All model parameters will be estimated by RAxML\n");
	      printf( "ML estimate of %d per site rate categories\n", adef->categories);      
	      printf( "Likelihood of final tree will be evaluated and optimized under GTR+GAMMA\n");
	      printf( "GTR+GAMMA Model parameters will be estimated up to an accuracy of %2.10f Log Likelihood units\n", adef->likelihoodEpsilon);
	      fprintf(infoFile, "GTR mixed model of nucleotide substitution\n"); 
	      fprintf(infoFile, "All model parameters will be estimated by RAxML\n");
	      fprintf(infoFile, "ML estimate of %d per site rate categories\n", adef->categories);
	      fprintf(infoFile, "Likelihood of final tree will be evaluated and optimized under GTR+GAMMA\n");
	      fprintf(infoFile, "GTR+GAMMA Model parameters will be estimated up to an accuracy of %2.10f Log Likelihood units\n", adef->likelihoodEpsilon);
	    }
	  else
	    {
	      printf( "GTR approximation of nucleotide substitution\n"); 
	      printf( "All model parameters will be estimated by RAxML\n");
	      printf( "ML estimate of %d per site rate categories\n", adef->categories);   
	      printf( "WARNING: GTR+CAT likelihood values are numerically instable!\n");
	      printf( "WARNING: For ML-based comparisons of final trees use -m GTRMIX\n");
	      fprintf(infoFile, "GTR approximation of nucleotide substitution\n"); 
	      fprintf(infoFile, "All model parameters will be estimated by RAxML\n");
	      fprintf(infoFile, "ML estimate of %d per site rate categories\n", adef->categories);
	      fprintf(infoFile, "WARNING: GTR+CAT likelihood values are numerically instable!\n");
	      fprintf(infoFile, "WARNING: For ML-based comparisons of final trees use -m GTRMIX\n");
	    }
	  break;   
	case M_GTRGAMMA:
	  printf( "GTR model of nucleotide substitution\n");  
	  printf( "All model parameters will be estimated by RAxML\n");  
	  printf( "GAMMA model of rate heteorgeneity, ML estimate of alpha-parameter\n");
	  printf( "GTR+GAMMA Model parameters will be estimated up to an accuracy of %2.10f Log Likelihood units\n", adef->likelihoodEpsilon);
	  fprintf(infoFile, "GTR model of nucleotide substitution\n");  
	  fprintf(infoFile, "All model parameters will be estimated by RAxML\n");  
	  fprintf(infoFile, "GAMMA model of rate heterogeneity, ML estimate of alpha-parameter\n");
	  fprintf(infoFile, "GTR+GAMMA Model parameters will be estimated up to an accuracy of %2.10f Log Likelihood units\n", adef->likelihoodEpsilon);
	  break;
	}
      
      if(adef->boot == 0)
	{
	  printf("Empirical Base Frequencies:\n");
	  fprintf(infoFile, "Empirical Base Frequencies:\n");
	  
	  if(tr->NumberOfModels == 1)
	    {
	      printf("pi(A): %f ",   tr->frequencies[0]);
	      printf("pi(C): %f ",   tr->frequencies[1]);
	      printf("pi(G): %f ",   tr->frequencies[2]);
	      printf("pi(T): %f \n", tr->frequencies[3]);                         
	      fprintf(infoFile, "pi(A): %f ",   tr->frequencies[0]);
	      fprintf(infoFile, "pi(C): %f ",   tr->frequencies[1]);
	      fprintf(infoFile, "pi(G): %f ",   tr->frequencies[2]);
	      fprintf(infoFile, "pi(T): %f \n", tr->frequencies[3]);  
	    }
	  else
	    {
	      int model;
	      
	      for(model = 0; model < tr->NumberOfModels; model++)
		{
		  printf("Partition %d: pi(A): %f ",   model, tr->frequencies[model * 4]);
		  printf("pi(C): %f ",   tr->frequencies[model * 4 + 1]);
		  printf("pi(G): %f ",   tr->frequencies[model * 4 + 2]);
		  printf("pi(T): %f \n", tr->frequencies[model * 4 + 3]);                         
		  fprintf(infoFile, "Partition %d: pi(A): %f ",   model, tr->frequencies[model *4]);
		  fprintf(infoFile, "pi(C): %f ",   tr->frequencies[model * 4 + 1]);
		  fprintf(infoFile, "pi(G): %f ",   tr->frequencies[model * 4 + 2]);
		  fprintf(infoFile, "pi(T): %f \n", tr->frequencies[model * 4 + 3]);
		}
	      
	    }
	}
      
      printf("\n\n");
      fprintf(infoFile, "\n\n");
      
      fprintf(infoFile,"RAxML was called as follows:\n\n");
      for(i = 0; i < argc; i++)
	fprintf(infoFile,"%s ", argv[i]);
      fprintf(infoFile,"\n\n\n");  
      
      fclose(infoFile);
    }
}

void printResult(tree *tr, analdef *adef)
{
  FILE *logFile;
  char temporaryFileName[1024] = "", treeID[64] = "";

  strcpy(temporaryFileName, resultFileName);

  switch(adef->mode)
    {
    case TREE_EVALUATION:	 
      break;
    case RELL_BOOTSTRAP:
      break;
    case BIG_RAPID_MODE:
      if(!adef->boot)
	{
	  if(adef->multipleRuns > 1)
	    {	  	 	  
	      sprintf(treeID, "%d", tr->treeID);	  	  
	      strcat(temporaryFileName, ".RUN.");
	      strcat(temporaryFileName, treeID);	  	 	      	
	    }
	       	      	     
	  Tree2String(tr->tree_string, tr, tr->start->back, TRUE, TRUE, FALSE, FALSE);

	  logFile = fopen(temporaryFileName, "w");
	  fprintf(logFile, "%s", tr->tree_string);
	  fclose(logFile);      
	}
      break;       
    default:
      printf("FATAL ERROR call to printResult from undefined STATE %d\n", adef->mode);
      exit(-1);
      break;
    }
}

void printBootstrapResult(tree *tr, analdef *adef)
{
  if(processID == 0)
    {
      FILE *logFile;
      
      if(adef->mode == BIG_RAPID_MODE && adef->boot)
	{
#ifndef PARALLEL
	  Tree2String(tr->tree_string, tr, tr->start->back, FALSE, TRUE, FALSE, FALSE);
#endif
	  logFile = fopen(bootstrapFileName, "a");
	  fprintf(logFile, "%s", tr->tree_string);
	  fclose(logFile);     
	}
      else
	{
	  printf("FATAL ERROR %d\n");
	  exit(-1);	 
	}
    }
}

void printRELLResult(tree *tr, analdef *adef)
{
  if(processID == 0)
    {
      FILE *logFile;
      
      if(adef->mode == RELL_BOOTSTRAP || adef->mode == CALC_BIPARTITIONS)
	{
	  Tree2String(tr->tree_string, tr, tr->start->back, TRUE, TRUE, FALSE, TRUE);
	  logFile = fopen(rellBootFileName, "a");
	  fprintf(logFile, "%s", tr->tree_string);
	  fclose(logFile);     
	}
      else
	{
	  printf("FATAL ERROR %d\n");
	  exit(-1);	 
	}
    }
}

void printLog(tree *tr, analdef *adef)
{
  FILE *logFile;
  char temporaryFileName[1024] = "", checkPoints[1024] = "", treeID[64] = "";
  double lh, t;
  
  lh = tr->likelihood;
  t = gettime() - masterTime;

  strcpy(temporaryFileName, logFileName);
  strcpy(checkPoints,       checkpointFileName);

  switch(adef->mode)
    {
    case TREE_EVALUATION:	 
      logFile = fopen(temporaryFileName, "a");

      printf("%f %f\n", t, lh);
      fprintf(logFile, "%f %f\n", t, lh);

      fclose(logFile);     
      break;
    case RELL_BOOTSTRAP:
      /* NOTHING PRINTED SO FAR */
      break;
    case BIG_RAPID_MODE:
      if(adef->boot)
	{
	  /* testing only printf("%f %f\n", t, lh);*/
	  /* NOTHING PRINTED so far */
	}
      else
	{
	  if(adef->multipleRuns > 1)
	    {	  	 	  
	      sprintf(treeID, "%d", tr->treeID);	  	  
	      strcat(temporaryFileName, ".RUN.");
	      strcat(temporaryFileName, treeID);	  	 
	      
	      strcat(checkPoints, ".RUN.");
	      strcat(checkPoints, treeID);	      	      
	    }


	  if(!adef->checkpoints)
	    {
	      logFile = fopen(temporaryFileName, "a");
#ifndef PARALLEL	      
	      printf("%f %f\n", t, lh);
#endif
	      fprintf(logFile, "%f %f\n", t, lh);
	      
	      fclose(logFile);
	    }
	  else
	    {
	      logFile = fopen(temporaryFileName, "a");
#ifndef PARALLEL	      
	      printf("%f %f %d\n", t, lh, tr->checkPointCounter);
#endif
	      fprintf(logFile, "%f %f %d\n", t, lh, tr->checkPointCounter);
	      
	      fclose(logFile);
	      
	      strcat(checkPoints, ".");

	      sprintf(treeID, "%d", tr->checkPointCounter);
	      strcat(checkPoints, treeID);
	      Tree2String(tr->tree_string, tr, tr->start->back, TRUE, TRUE, FALSE, FALSE);

	      logFile = fopen(checkPoints, "a");
	      fprintf(logFile, "%s", tr->tree_string);
	      fclose(logFile);

	      tr->checkPointCounter++;
	    }
	}
      break;       
    default:
      printf("FATAL ERROR call to printLog from undefined STATE %d\n", adef->mode);
      exit(-1);
      break;
    }
}



void printStartingTree(tree *tr, analdef *adef)
{
  if(adef->boot)
    {    
      /* not printing starting trees for bootstrap */
    }
  else
    {
      FILE *treeFile;
      char temporaryFileName[1024] = "", treeID[64] = "";

      Tree2String(tr->tree_string, tr, tr->start->back, FALSE, TRUE, FALSE, FALSE);

      if(adef->randomStartingTree)	    
	strcpy(temporaryFileName, randomFileName);	    
      else
	strcpy(temporaryFileName, permFileName);

      if(adef->multipleRuns > 1)
	{	  	 	  
	  sprintf(treeID, "%d", tr->treeID);	  	  
	  strcat(temporaryFileName, ".RUN.");
	  strcat(temporaryFileName, treeID);	  	 
	}
     	  	 
      treeFile = fopen(temporaryFileName, "a");	
      fprintf(treeFile, "%s", tr->tree_string);
      fclose(treeFile);	    	
    }
}

void writeInfoFile(analdef *adef, tree *tr, double t)
{  
  if(processID == 0)
    {
      FILE *infoFile = fopen(infoFileName, "a");
      
      switch(adef->mode)
	{
	case TREE_EVALUATION:	 	  
	  break;
	case  RELL_BOOTSTRAP:	  
	  break;
	case BIG_RAPID_MODE:
	  if(adef->boot)
	    {
	      if(!adef->initialSet)	       
		fprintf(infoFile, "Bootstrap[%d]: Time %f bootstrap likelihood %f, best rearrangement setting %d\n", tr->treeID, t, tr->likelihood,  adef->bestTrav);		
	      else		
		fprintf(infoFile, "Bootstrap[%d]: Time %f bootstrap likelihood %f\n", tr->treeID, t, tr->likelihood);		
	    }
	  else
	    {
	      if(adef->model == M_GTRCAT && !adef->useMixedModel)
		{		  
		  if(!adef->initialSet)		   
		    fprintf(infoFile, "Inference[%d]: Time %f GTRCAT-likelihood %f, best rearrangement setting %d\n", tr->treeID, t, tr->likelihood,  adef->bestTrav);
		  else		  
		    fprintf(infoFile, "Inference[%d]: Time %f GTRCAT-likelihood %f\n", tr->treeID, t, tr->likelihood);		    
		}
	      else
		{	
		  int model;

		  if(!adef->initialSet)		    		     	  
		    fprintf(infoFile, "Inference[%d]: Time %f GTRGAMMA-likelihood %f, best rearrangement setting %d, alpha ", 
			    tr->treeID, t, tr->likelihood,  adef->bestTrav);
		  else		  
		    fprintf(infoFile, "Inference[%d]: Time %f GTRGAMMA-likelihood %f, alpha ", tr->treeID, t, tr->likelihood);		    		  

		  for(model = 0; model < tr->NumberOfModels; model++)		    		    
		    fprintf(infoFile, "%f ", tr->alphas[model]);
		    		  
		  fprintf(infoFile, "\n");
		}
	    }
	  break;
	}

      fclose(infoFile);
    }
}

void finalizeInfoFile(tree *tr, analdef *adef)
{
  if(processID == 0)
    {
      FILE *infoFile;
      double t;

      t = gettime() - masterTime;

      switch(adef->mode)
	{       
	case TREE_EVALUATION :
	  printf("\n\nOverall Time for Tree Evaluation %f\n", t);
	  printf("Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	  if(tr->NumberOfModels == 1)
	    printf("Final alpha shape parameter: %f\n", tr->alphas[0]);
	  else
	    {
	      int model;
	      for(model = 0; model < tr->NumberOfModels; model++)
		printf("Partition %d: final alpha shape parameter: %f\n", model, tr->alphas[model]);	      
	    }
	  
	  infoFile = fopen(infoFileName, "a");
	  
	  fprintf(infoFile, "\n\nOverall Time for Tree Evaluation %f\n", t);
	  fprintf(infoFile, "Final GTR+GAMMA  likelihood: %f\n", tr->likelihood);
	  if(tr->NumberOfModels == 1)
	    fprintf(infoFile, "Final alpha shape parameter: %f\n", tr->alphas[0]);		
	  else
	    {
	      int model;
	      for(model = 0; model < tr->NumberOfModels; model++)
		fprintf(infoFile, "Partition %d: final alpha shape parameter: %f\n", model, tr->alphas[model]);
	    
	    }	
	         	 
	  fclose(infoFile);
	  break;  
	case  BIG_RAPID_MODE:
	  if(adef->boot)
	    {
	      printf("\n\nOverall Time for %d Bootstraps %f\n", adef->multipleRuns, t);
	      printf("\n\nAverage Time per Bootstrap %f\n", (double)(t/((double)adef->multipleRuns)));
	      printf("All %d bootstrapped trees written to: %s\n", adef->multipleRuns, bootstrapFileName);
	      infoFile = fopen(infoFileName, "a");
	      fprintf(infoFile, "\n\nOverall Time for %d Bootstraps %f\n", adef->multipleRuns, t);
	      fprintf(infoFile, "Average Time per Bootstrap %f\n", (double)(t/((double)adef->multipleRuns)));	     
	      fprintf(infoFile, "\n\nAll %d bootstrapped trees written to: %s\n", adef->multipleRuns, bootstrapFileName);
	      fclose(infoFile);
	    }
	  else
	    {
	      if(adef->multipleRuns > 1)
		{
		  double avgLH = 0;
		  double bestLH = unlikely;
		  int i, bestI  = 0;
		  
		  for(i = 0; i < adef->multipleRuns; i++)
		    {     
		      avgLH   += tr->likelihoods[i];
		      if(tr->likelihoods[i] > bestLH)
			{
			  bestLH = tr->likelihoods[i];
			  bestI  = i;
			}
		    }
		  avgLH /= ((double)adef->multipleRuns);

		  printf("\n\nOverall Time for %d Inferences %f\n", adef->multipleRuns, t);
		  printf("Average Time per Inference %f\n", (double)(t/((double)adef->multipleRuns)));		 
		  printf("Average Likelihood   : %f\n", avgLH);
		  printf("\n");
		  printf("Best Likelihood in run number %d: likelihood %f\n\n", bestI, bestLH);

		  if(adef->checkpoints)   
		    printf("Checkpoints written to:                 %s.RUN.%d.* to %d.*\n", checkpointFileName, 0, adef->multipleRuns - 1);  
		  if(!adef->restart)
		    {
		      if(adef->randomStartingTree)
			printf("Random starting trees written to:       %s.RUN.%d to %d\n", randomFileName, 0, adef->multipleRuns - 1);
		      else
			printf("Parsimony starting trees written to:    %s.RUN.%d to %d\n", permFileName, 0, adef->multipleRuns - 1);   	            
		    }					  
		  printf("Final trees written to:                 %s.RUN.%d to %d\n", resultFileName,  0, adef->multipleRuns - 1);  
		  printf("Execution Log Files written to:         %s.RUN.%d to %d\n", logFileName, 0, adef->multipleRuns - 1);   
		  printf("Execution information file written to:  %s\n", infoFileName);


		  infoFile = fopen(infoFileName, "a");

		  fprintf(infoFile, "\n\nOverall Time for %d Inferences %f\n", adef->multipleRuns, t);
		  fprintf(infoFile, "Average Time per Inference %f\n", (double)(t/((double)adef->multipleRuns)));		 
		  fprintf(infoFile, "Average Likelihood   : %f\n", avgLH);
		  fprintf(infoFile, "\n");
		  fprintf(infoFile, "Best Likelihood in run number %d: likelihood %f\n\n", bestI, bestLH); 
		  if(adef->checkpoints)   
		    fprintf(infoFile, "Checkpoints written to:                %s.RUN.%d.* to %d.*\n", checkpointFileName, 0, adef->multipleRuns - 1);  
		  if(!adef->restart)
		    {
		      if(adef->randomStartingTree)
			fprintf(infoFile, "Random starting trees written to:      %s.RUN.%d to %d\n", randomFileName, 0, adef->multipleRuns - 1);
		      else
			fprintf(infoFile, "Parsimony starting trees written to:   %s.RUN.%d to %d\n", permFileName, 0, adef->multipleRuns - 1);   	            
		    }					  
		  fprintf(infoFile, "Final trees written to:                %s.RUN.%d to %d\n", resultFileName,  0, adef->multipleRuns - 1);  
		  fprintf(infoFile, "Execution Log Files written to:        %s.RUN.%d to %d\n", logFileName, 0, adef->multipleRuns - 1);   
		  fprintf(infoFile, "Execution information file written to: %s\n", infoFileName);

		  fclose(infoFile);		  		 		   
		}
	      else
		{
		  printf("\n\nOverall Time for 1 Inference %f\n", t);		  
		  printf("Likelihood   : %f\n", tr->likelihood);
		  printf("\n\n");	     

		  if(adef->checkpoints)   
		  printf("Checkpoints written to:                %s.*\n", checkpointFileName);  
		  if(!adef->restart)
		    {
		      if(adef->randomStartingTree)
			printf("Random starting tree written to:       %s\n", randomFileName);
		      else
			printf("Parsimony starting tree written to:    %s\n", permFileName);   	            
		    }					  
		  printf("Final tree written to:                 %s\n", resultFileName);  
		  printf("Execution Log File written to:         %s\n", logFileName);   
		  printf("Execution information file written to: %s\n",infoFileName);

		  infoFile = fopen(infoFileName, "a");
		  
		  fprintf(infoFile, "\n\nOverall Time for 1 Inference %f\n", t);		  
		  fprintf(infoFile, "Likelihood   : %f\n", tr->likelihood);
		  fprintf(infoFile, "\n\n");

		  if(adef->checkpoints)   
		    fprintf(infoFile, "Checkpoints written to:                %s.*\n", checkpointFileName);  
		  if(!adef->restart)
		    {
		      if(adef->randomStartingTree)
			fprintf(infoFile, "Random starting tree written to:       %s\n", randomFileName);
		      else
			fprintf(infoFile, "Parsimony starting tree written to:    %s\n", permFileName);   	            
		    }					  
		  fprintf(infoFile, "Final tree written to:                 %s\n", resultFileName);  
		  fprintf(infoFile, "Execution Log File written to:         %s\n", logFileName);   
		  fprintf(infoFile, "Execution information file written to: %s\n",infoFileName);
		  
		  fclose(infoFile);		  
		}
	    }
	    
	  break;
	case RELL_BOOTSTRAP:

	  printf("\n\n%d alternative topologies have been relled\n", numberOfRelledTopologies);	 
	  printf("Overall Time for RELL bootstrap %f\n", t);
	  printf("RELL-Bootstrap Information written to file %s\n",rellInfoFileName);
	  printf("RELL-Bootstrap Tree written to file %s\n", rellBootFileName);

	  infoFile = fopen(infoFileName, "a");
	   
	  fprintf(infoFile, "\n\n%d alternative topologies have been relled\n", numberOfRelledTopologies);
	  fprintf(infoFile, "Overall Time for RELL bootstrap %f\n", t);
	  fprintf(infoFile, "RELL-Bootstrap Information written to file %s\n",rellInfoFileName);
	  fprintf(infoFile, "RELL-Bootstrap Tree written to file %s\n", rellBootFileName);

	  fclose(infoFile);
	  break;
	case CALC_BIPARTITIONS:
	  printf("\n\nOverall Time for Computation of Bipartitions %f\n", t);
	  printf("Tree with bipartition information written to file %s\n", rellBootFileName);

	  infoFile = fopen(infoFileName, "a");

	  fprintf(infoFile, "\n\nOverall Time for Computation of Bipartitions %f\n", t);
	  fprintf(infoFile, "Tree with bipartition information written to file %s\n", rellBootFileName);

	  fclose(infoFile);
	  break;
	}

    }

}

/********************PRINTING various INFO **************************************/

/************************************************************************************/

main (int argc, char *argv[]) 
{   
  rawdata      *rdta;
  cruncheddata *cdta;
  tree         *tr;         
  analdef      *adef;
  
  masterTime = gettime();            

  adef = (analdef *)malloc(sizeof(analdef));
  rdta = (rawdata *)malloc(sizeof(rawdata));
  cdta = (cruncheddata *)malloc(sizeof(cruncheddata));     
  tr   = (tree *)malloc(sizeof(tree));

  initAdef(adef);
  get_args(argc,argv,(boolean)1, adef, rdta);              
  /* TODO alle fehlerabfragen in get_args und dort exit */

#ifdef PARALLEL
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &processID);    
  MPI_Comm_size(MPI_COMM_WORLD, &numOfWorkers);      
#endif  
#ifndef PARALLEL
  processID = 0;
#endif
  
  readData(FALSE, adef, rdta, cdta, tr);
  
  makeweights(adef, rdta, cdta, tr);
  makevalues(rdta, cdta, tr, adef);
  linkdata2tree(rdta, cdta, tr);    
  initModel(tr, rdta, cdta, adef);                                     
  
  makeFileNames(tr, adef, argc, argv);        
  printModelAndProgramInfo(tr, adef, argc, argv); 
  
  switch(adef->mode)
    {
    case TREE_EVALUATION:    
      getStartingTree(tr, adef);
      modOpt(tr, adef);
      printLog(tr, adef);   
      break;
      /*case RELL_BOOTSTRAP:    
      rellBootstrap(tr, adef);  
      break;
      case CALC_BIPARTITIONS:    
      calcBipartitions(tr, adef);   
      break;*/
    case BIG_RAPID_MODE:
      if(adef->boot)    
	doBootstrap(tr, adef, rdta, cdta);
      else  
	doInference(tr, adef, rdta, cdta);          
      break;
    }

  finalizeInfoFile(tr, adef);

#ifdef PARALLEL
  MPI_Finalize();
#endif

  return 1;
}
