/*  RAxML-VI-HPC (version 2.0) a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright June 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  stamatak@ics.forth.gr
 *
 *  When publishing work that is based on the results from RAxML-HPC please cite:
 *
 *  Alexandros Stamatakis, Thomas Ludwig, and Harald Meier: "RAxML-III: A Fast Program for Maximum Likelihood-based Inference of Large Phylogenetic Trees". 
 *  In Bioinformatics 21(4):456-463.
 */


#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h> 
#include <math.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>


#include "axml.h"

extern int optimizeRatesInvocations;
extern int optimizeRateCategoryInvocations;
extern int optimizeAlphaInvocations;

void baseFrequenciesGTR(rawdata *rdta, cruncheddata *cdta, tree *tr, analdef *adef)
{
  double  sum, suma, sumc, sumg, sumt, wj, fa, fc, fg, ft, freqa, freqc, freqg, freqt;
  int     i, j, k, code;
  char  *yptr;

  if(!adef->useMultipleModel)
    {
      freqa = 0.25;
      freqc = 0.25;
      freqg = 0.25;
      freqt = 0.25;
      for (k = 1; k <= 8; k++) 
	{
	  suma = 0.0;
	  sumc = 0.0;
	  sumg = 0.0;
	  sumt = 0.0;
	  for (i = 0; i < rdta->numsp; i++) 
	    {
	      yptr = rdta->y[i];
	      for (j = 0; j < cdta->endsite; j++) 
		{
		  code = *yptr++;
		  fa = freqa * ( code       & 1);
		  fc = freqc * ((code >> 1) & 1);
		  fg = freqg * ((code >> 2) & 1);
		  ft = freqt * ((code >> 3) & 1);
		  wj = cdta->aliaswgt[j] / (fa + fc + fg + ft);
		  suma += wj * fa;
		  sumc += wj * fc;
		  sumg += wj * fg;
		  sumt += wj * ft;
		}
	    }
	  sum = suma + sumc + sumg + sumt;
	  freqa = suma / sum;
	  freqc = sumc / sum;
	  freqg = sumg / sum;
	  freqt = sumt / sum;
	}
      
      tr->frequencies[0] = freqa;
      tr->frequencies[1] = freqc;
      tr->frequencies[2] = freqg;
      tr->frequencies[3] = freqt;
    }
  else
    {
      int model;
      int lower;
      int upper;

      for(model = 0; model < tr->NumberOfModels; model++)
	{
	  lower = tr->modelIndices[model][0];
	  upper = tr->modelIndices[model][1];

	  freqa = 0.25;
	  freqc = 0.25;
	  freqg = 0.25;
	  freqt = 0.25;
	  for (k = 1; k <= 8; k++) 
	    {
	      suma = 0.0;
	      sumc = 0.0;
	      sumg = 0.0;
	      sumt = 0.0;
	      for (i = 0; i < rdta->numsp; i++) 
		{
		  yptr = &(rdta->y[i][lower]);
		  for (j = lower; j < upper; j++) 
		    {
		      code = *yptr++;
		      fa = freqa * ( code       & 1);
		      fc = freqc * ((code >> 1) & 1);
		      fg = freqg * ((code >> 2) & 1);
		      ft = freqt * ((code >> 3) & 1);
		      wj = cdta->aliaswgt[j] / (fa + fc + fg + ft);
		      suma += wj * fa;
		      sumc += wj * fc;
		      sumg += wj * fg;
		      sumt += wj * ft;
		    }
		}
	      sum = suma + sumc + sumg + sumt;
	      freqa = suma / sum;
	      freqc = sumc / sum;
	      freqg = sumg / sum;
	      freqt = sumt / sum;
	    }
      
	  tr->frequencies[model * 4] = freqa;
	  tr->frequencies[model * 4 + 1] = freqc;
	  tr->frequencies[model * 4 + 2] = freqg;
	  tr->frequencies[model * 4 + 3] = freqt;

	  /*printf("Ferquencies Partition %d; %f %f %f %f\n", model, freqa, freqc, freqg, freqt);*/
	}

    }      
 
  return;
}



void initReversibleGTR(rawdata *rdta, cruncheddata *cdta, tree *tr, analdef *adef, int model)
{ 
  double r[4][4], a[4][4], *initialRates = &(tr->initialRates[model * 5]), 
    f[4], e[4], d[4];
  int i, j, k, m, jj, code;
  double invfreq[4], EIGN[4], EIGV[4][4]; 
  double *eptr;
         
  f[0] = tr->frequencies[model * 4];
  f[1] = tr->frequencies[model * 4 + 1];
  f[2] = tr->frequencies[model * 4 + 2];
  f[3] = tr->frequencies[model * 4 + 3];
  
  i = 0;
  for (j = 0; j < 2; j++)
    for (k = j+1; k< 4; k++)
      r[j][k] = initialRates[i++];
  r[2][3] = 1;
  for (j = 0; j < 4; j++) 
    {
      r[j][j] = 0;
      for (k = 0; k < j; k++)
	r[j][k] = r[k][j];
    }
  
  /* TODO MULT SOSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS*/
  tr->fracchanges[model] = 0;
  tr->fracchange = 0;

  for (j = 0; j<4; j++)
    for (k = 0; k<4; k++)
      tr->fracchanges[model] += f[j] * r[j][k] * f[k];

  if(tr->NumberOfModels == 1)    
    tr->fracchange = tr->fracchanges[model];
    
  /*TODO MULT SOSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS*/
  m = 0;
  
  for(i=0; i<4; i++) 
    a[i][i] = 0;
  
  for(i=0; i<4; i++) 
    {
      for(j=i+1; j<4; j++) 
	{
	  double factor = ((m>=5) ? 1 : initialRates[m++]);
	  a[i][j] = a[j][i] = factor * sqrt( f[i] * f[j]);
	  a[i][i] -= factor * f[j];
	  a[j][j] -= factor * f[i];
	}
    }
  
  tred2((double *)a,4,4,d,e);       
  tqli(d, e, 4 , 4, (double *)a);
  
  for(i=0; i<4; i++)     
    for(j=0; j<4; j++)       
      a[i][j] *= sqrt(f[j]);
  
  for (i=0; i<4; i++)
    {	  
      if (d[i] > -1e-8) 
	{	      
	  if (i != 0) 
	    {		    
	      double tmp = d[i], sum=0;
	      d[i] = d[0];
	      d[0] = tmp;
	      for (j=0; j < 4; j++) 
		{
		  tmp = a[i][j];
		  a[i][j] = a[0][j];
		  sum += (a[0][j] = tmp);
		}
	      for (j=0; j < 4; j++) 
		a[0][j] /= sum;
	    }
	  break;
	}
    }
  
  for (i=0; i< 4; i++) 
    {
      EIGN[i] = -d[i];
      for (j=0; j<4; j++)
	EIGV[i][j] = a[j][i];
      invfreq[i] = 1 / EIGV[i][0]; 
    }                   
  
  tr->EIGN[model * 3] = EIGN[1];
  tr->EIGN[model * 3 + 1] = EIGN[2];
  tr->EIGN[model * 3 + 2] = EIGN[3];
  
  eptr = &(tr->EV[model * 16]);
  
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)      
      *eptr++ = EIGV[i][j];	      
  
      
  tr->EI[model * 12] = tr->EV[model * 16 + 1] * invfreq[0];
  tr->EI[model * 12 + 1] = tr->EV[model * 16 + 2] * invfreq[0];
  tr->EI[model * 12 + 2] = tr->EV[model * 16 + 3] * invfreq[0];
  
  tr->EI[model * 12 + 3] = tr->EV[model * 16 + 5] * invfreq[1];
  tr->EI[model * 12 + 4] = tr->EV[model * 16 + 6] * invfreq[1];
  tr->EI[model * 12 + 5] = tr->EV[model * 16 + 7] * invfreq[1];
  
  tr->EI[model * 12 + 6] = tr->EV[model * 16 + 9]  * invfreq[2];
  tr->EI[model * 12 + 7] = tr->EV[model * 16 + 10] * invfreq[2];
  tr->EI[model * 12 + 8] = tr->EV[model * 16 + 11] * invfreq[2];
  
  tr->EI[model * 12 + 9]  = tr->EV[model * 16 + 13] * invfreq[3];
  tr->EI[model * 12 + 10] = tr->EV[model * 16 + 14] * invfreq[3];
  tr->EI[model * 12 + 11] = tr->EV[model * 16 + 15] * invfreq[3];	     
  
     
  for(i=0; i < 16; i++)
    {
      code = i;
      tr->gtrTip[model * 16 + i].a = 0;
      tr->gtrTip[model * 16 + i].c = 0;
      tr->gtrTip[model * 16 + i].g = 0;
      tr->gtrTip[model * 16 + i].t = 0;
      for (j = 0; j < 4; j++) 
	{
	  if ((code >> j) & 1) 
	    {
	      int jj = "0213"[j] - '0';		
	      tr->gtrTip[model * 16 + i].a += EIGV[jj][0];
	      tr->gtrTip[model * 16 + i].c += EIGV[jj][1];
	      tr->gtrTip[model * 16 + i].g += EIGV[jj][2];
	      tr->gtrTip[model * 16 + i].t += EIGV[jj][3];
	    }
	}
     
    }
}


double LnGamma (double alpha)
{
/* returns ln(gamma(alpha)) for alpha>0, accurate to 10 decimal places.  
   Stirling's formula is used for the central polynomial part of the procedure.
   Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
   Communications of the Association for Computing Machinery, 9:684
*/
  double x, f, z, result;

  x = alpha;
  f = 0.0;
  
  if ( x < 7.0) 
     {
       f = 1.0;  
       z = alpha - 1.0;
      
       while ((z = z + 1.0) < 7.0)  
	 {	  
	   f *= z;
	 }
       x = z;   
     
       if(f == 0.0)
	 {
	   printf("FATAL ERROR in LnGamma\n");
	   exit(-1);
	 }
       f=-log(f);
     }
   
   z = 1/(x*x);
   
   result = f + (x-0.5)*log(x) - x + .918938533204673 
	  + (((-.000595238095238*z+.000793650793651)*z-.002777777777778)*z
	       +.083333333333333)/x;  

   return result;
}



double IncompleteGamma (double x, double alpha, double ln_gamma_alpha)
{
/* returns the incomplete gamma ratio I(x,alpha) where x is the upper 
	   limit of the integration and alpha is the shape parameter.
   returns (-1) if in error
   ln_gamma_alpha = ln(Gamma(alpha)), is almost redundant.
   (1) series expansion     if (alpha>x || x<=1)
   (2) continued fraction   otherwise
   RATNEST FORTRAN by
   Bhattacharjee GP (1970) The incomplete gamma integral.  Applied Statistics,
   19: 285-287 (AS32)
*/
   int i;
   double p=alpha, g=ln_gamma_alpha;
   double accurate=1e-8, overflow=1e30;
   double factor, gin=0, rn=0, a=0,b=0,an=0,dif=0, term=0, pn[6];


   if (x==0) return (0);
   if (x<0 || p<=0) return (-1);

   
   factor=exp(p*log(x)-x-g);   
   if (x>1 && x>=p) goto l30;
   /* (1) series expansion */
   gin=1;  term=1;  rn=p;
 l20:
   rn++;
   term*=x/rn;   gin+=term;

   if (term > accurate) goto l20;
   gin*=factor/p;
   goto l50;
 l30:  
   /* (2) continued fraction */
   a=1-p;   b=a+x+1;  term=0;
   pn[0]=1;  pn[1]=x;  pn[2]=x+1;  pn[3]=x*b;
   gin=pn[2]/pn[3];   
 l32:  
   a++;  
   b+=2;  
   term++;   
   an=a*term;
   for (i=0; i<2; i++) 
     pn[i+4]=b*pn[i+2]-an*pn[i];
   if (pn[5] == 0) goto l35;
   rn=pn[4]/pn[5];   
   dif=fabs(gin-rn);  
   if (dif>accurate) goto l34;
   if (dif<=accurate*rn) goto l42;
 l34:   
   gin=rn;
 l35: 
   for (i=0; i<4; i++) 
     pn[i]=pn[i+2];
   if (fabs(pn[4]) < overflow)            
     goto l32;        
   
   for (i=0; i<4; i++) 
     pn[i]/=overflow;

   
   goto l32;
 l42:  
   gin=1-factor*gin;

 l50: 
   return (gin);
}




double PointNormal (double prob)
{
/* returns z so that Prob{x<z}=prob where x ~ N(0,1) and (1e-12)<prob<1-(1e-12)
   returns (-9999) if in error
   Odeh RE & Evans JO (1974) The percentage points of the normal distribution.
   Applied Statistics 22: 96-97 (AS70)

   Newer methods:
     Wichura MJ (1988) Algorithm AS 241: the percentage points of the
       normal distribution.  37: 477-484.
     Beasley JD & Springer SG  (1977).  Algorithm AS 111: the percentage 
       points of the normal distribution.  26: 118-121.

*/
   double a0=-.322232431088, a1=-1, a2=-.342242088547, a3=-.0204231210245;
   double a4=-.453642210148e-4, b0=.0993484626060, b1=.588581570495;
   double b2=.531103462366, b3=.103537752850, b4=.0038560700634;
   double y, z=0, p=prob, p1;

   p1 = (p<0.5 ? p : 1-p);
   if (p1<1e-20) return (-9999);

   y = sqrt (log(1/(p1*p1)));   
   z = y + ((((y*a4+a3)*y+a2)*y+a1)*y+a0) / ((((y*b4+b3)*y+b2)*y+b1)*y+b0);
   return (p<0.5 ? -z : z);
}


double PointChi2 (double prob, double v)
{
/* returns z so that Prob{x<z}=prob where x is Chi2 distributed with df=v
   returns -1 if in error.   0.000002<prob<0.999998
   RATNEST FORTRAN by
       Best DJ & Roberts DE (1975) The percentage points of the 
       Chi2 distribution.  Applied Statistics 24: 385-388.  (AS91)
   Converted into C by Ziheng Yang, Oct. 1993.
*/
   double e=.5e-6, aa=.6931471805, p=prob, g;
   double xx, c, ch, a=0,q=0,p1=0,p2=0,t=0,x=0,b=0,s1,s2,s3,s4,s5,s6;
  
   if (p<.000002 || p>.999998 || v<=0) return (-1);
  
   g = LnGamma(v/2);
   
   xx=v/2;   c=xx-1;
   if (v >= -1.24*log(p)) goto l1;

   ch=pow((p*xx*exp(g+xx*aa)), 1/xx);
   if (ch-e<0) return (ch);
   goto l4;
l1:
   if (v>.32) goto l3;
   ch=0.4;   a=log(1-p);
l2:
   q=ch;  p1=1+ch*(4.67+ch);  p2=ch*(6.73+ch*(6.66+ch));
   t=-0.5+(4.67+2*ch)/p1 - (6.73+ch*(13.32+3*ch))/p2;
   ch-=(1-exp(a+g+.5*ch+c*aa)*p2/p1)/t;
   if (fabs(q/ch-1)-.01 <= 0) goto l4;
   else                       goto l2;
  
l3:    
   x=PointNormal (p);
   p1=0.222222/v;   ch=v*pow((x*sqrt(p1)+1-p1), 3.0);
   if (ch>2.2*v+6)  ch=-2*(log(1-p)-c*log(.5*ch)+g);
l4:
   q=ch;   p1=.5*ch;   
   if ((t=IncompleteGamma (p1, xx, g))< 0.0) 
     {
       printf ("IncompleteGamma \n");      
       return (-1);
     }
  
   p2=p-t;
   t=p2*exp(xx*aa+g+p1-c*log(ch));   
   b=t/ch;  a=0.5*t-b*c;

   s1=(210+a*(140+a*(105+a*(84+a*(70+60*a))))) / 420;
   s2=(420+a*(735+a*(966+a*(1141+1278*a))))/2520;
   s3=(210+a*(462+a*(707+932*a)))/2520;
   s4=(252+a*(672+1182*a)+c*(294+a*(889+1740*a)))/5040;
   s5=(84+264*a+c*(175+606*a))/2520;
   s6=(120+c*(346+127*c))/5040;
   ch+=t*(1+0.5*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))));
   if (fabs(q/ch-1) > e) goto l4;

   return (ch);
}

#define PointGamma(prob,alpha,beta)  PointChi2(prob,2.0*(alpha))/(2.0*(beta))




void makeGammaCats(tree *tr, int model)
{
  int i, K;
  double gap05, t, factor, lnga1, alfa, beta;
  double gammaProbs[4];

  K = 4;
  alfa = beta = tr->alphas[model];

  if(alfa < ALPHA_MIN)
    {
      printf("GAMMA_CATS alfa smaller ALPHA_MIN\n");
      exit(-1);
    }

  gap05 = 1.0 / (2.0 * K);
  factor = alfa / beta * K;
  
  if(K==1) 
     {
       printf("ERROR in makeGammaCats\n");
       exit(-1);
     }  

   lnga1=LnGamma(alfa+1);
   
   for (i=0; i<K-1; i++)
     gammaProbs[i]=PointGamma((i+1.0)/K, alfa, beta);
  
   for (i=0; i<K-1; i++)
     gammaProbs[i]=IncompleteGamma(gammaProbs[i]*beta, alfa+1, lnga1);

   tr->gammaRates[model * 4] = gammaProbs[0]*factor;

   tr->gammaRates[model * 4 + (K-1)] = (1 - gammaProbs[K-2])*factor;

   for (i=1; i<K-1; i++)  
     tr->gammaRates[model * 4 + i] = (gammaProbs[i] - gammaProbs[i-1])*factor;    
   
   return;  
}

void initModel(tree *tr, rawdata *rdta, cruncheddata *cdta, analdef *adef)
{  
  int model, i, j;
  double  temp, wtemp;

  if(!adef->useMultipleModel)
    tr->NumberOfModels = 1;
 
  optimizeRatesInvocations = 1;  
  optimizeRateCategoryInvocations = 1;  
  optimizeAlphaInvocations = 1;   

  for(model = 0; model < tr->NumberOfModels; model++)
    {
      tr->initialRates[model * 5] = 0.5;
      tr->initialRates[model * 5 + 1] =  0.5;
      tr->initialRates[model * 5 + 2] = 0.5;
      tr->initialRates[model * 5 + 3] = 0.5;
      tr->initialRates[model * 5 + 4] = 0.5;      
    }

  switch(adef->model)
    {       
    case M_GTRGAMMA:
      if(adef->useMultipleModel)
	{
	  newview = newviewGTRGAMMAMULT;
	  makenewz = makenewzGTRGAMMAMULT;
	  evaluate = evaluateGTRGAMMAMULT;

	  evaluatePartition = evaluateGTRGAMMAMULTPARTITION;
	  newviewPartition = newviewGTRGAMMAMULTPARTITION;
	}
      else
	{
	  newview = newviewGTRGAMMA;
	  makenewz = makenewzGTRGAMMA;
	  evaluate = evaluateGTRGAMMA;	  	 
	}      
      for(j = 0; j < tr->cdta->endsite; j++)
	tr->cdta->wr[j] = tr->cdta->aliaswgt[j];
      baseFrequenciesGTR(rdta, cdta, tr, adef);                 
      for(model = 0; model < tr->NumberOfModels; model++)
	{	
	  tr->alphas[model] = 1.0;
	  initReversibleGTR(rdta, cdta, tr, adef, model);     
	  makeGammaCats(tr, model);      
	}
      
      break;
    case M_GTRCAT:
      if(adef->useMultipleModel)
	{	
	  newview = newviewGTRCATMULT; 
	  makenewz = makenewzGTRCATMULT; 
	  evaluate = evaluateGTRCATMULT;

	  evaluatePartial = evaluatePartialGTRCATMULT;
	  
	  evaluatePartition = evaluateGTRCATMULTPARTITION;
	  newviewPartition = newviewPARTITIONGTRCATMULT;
	}
      else
	{      
	  newview = newviewGTRCAT; 
	  makenewz = makenewzGTRCAT;
	  evaluate = evaluateGTRCAT;
	  evaluatePartial = evaluatePartialGTRCAT;

	  evaluatePartition = evaluateGTRCATMULTPARTITION;
	  newviewPartition = newviewPARTITIONGTRCATMULT;
	}
      for (j = 0; j < tr->cdta->endsite; j++) 
	{
	  tr->cdta->patrat[j] = temp = 1.0;
	  tr->cdta->patratStored[j] = 1.0;
	  tr->cdta->rateCategory[j] = 0;
	  tr->cdta->wr[j]  = wtemp = temp * tr->cdta->aliaswgt[j];
	  tr->cdta->wr2[j] = temp * wtemp;
	}     
      baseFrequenciesGTR(rdta, cdta, tr,adef);       
      tr->NumberOfCategories = 1;    
      for(model = 0; model < tr->NumberOfModels; model++)
	initReversibleGTR(rdta, cdta, tr, adef, model);            
        
      break;
    }

  if(tr->NumberOfModels > 1)
    {
      tr->fracchange = 0;
      for(model = 0; model < tr->NumberOfModels; model++)
	{
	  tr->fracchange += tr->fracchanges[model];
	}
      tr->fracchange /= ((double)tr->NumberOfModels);
    }

}

